;;;
;;;  sample for xawscm and widget macro package.
;;;
;;;  @(#)xawedit.scm 1.0.0 94/02/15 (Shiro Kawai, kawai@sail.t.u-tokyo.ac.jp)
;;;

(require 'repl)
(require 'x11)
(require 'xt)
(require 'xevent)
(require 'xaw)

(if (not (memq 'posix *features*))
    (load (in-vicinity (implementation-vicinity) "posix")))
(if (not (memq 'i/o-extentions *features*))
    (load (in-vicinity (implementation-vicinity) "ioext")))
(load  "widget")

;; Global variables
(define *editing-file* #f)

;; MAIN-WINDOW
(define main-window
  (widget "main-window" xaw:paned
	  (:resources xt:n-width 600)
	  (:children menu-bar text-pane)))

;; MAIN-MENU
(define menu-bar
  (widget "menu-bar" xaw:box
	  (:resources xt:n-orientation "horizontal")
	  (:children
	   (widget "file-button" xaw:menu-button
		   (:resources xt:n-label "File"
			       xt:n-menu-name "file-menu")
		   (:init (lambda (self child)
			    (file-menu self))))
	   (widget "search-button" xaw:command
		   (:resources xt:n-label "Search")
		   (:callbacks xt:n-callback
			       (lambda (w)
				 ;; XXX Not implemened yet
				 (warn "Sorry, not implemented yet")))))))


;; FILE-MENU
(define file-menu
  (widget "file-menu" xaw:simple-menu
	   (:instantiation xt:create-popup-shell)
	   (:children
	     (widget "open" xaw:sme-bsb
		      (:resources xt:n-label "open")
		      (:callbacks xt:n-callback
				  (lambda (w)
				    (xt:popup (file-open-w toplevel-w) 2))
				  ))
	     (widget "save" xaw:sme-bsb
		      (:resources xt:n-label "save")
		      (:callbacks xt:n-callback
				  (lambda (w)
				    (if *editing-file*
					(if (xaw:ascii-save *editing-file*)
					    '()
					    (warn "Save Failed."))))))
	     (widget "interpreter" xaw:sme-bsb
		      (:resources xt:n-label "interpreter")
		      (:callbacks xt:n-callback
				  (lambda (w)
				    (display "### interpreter invoked")
				    (newline)
				    (display "    (repl:quit) to exit")
				    (newline)
				    (repl:top-level slib:eval))))
	     (widget "quit" xaw:sme-bsb
		      (:resources xt:n-label "quit")
		      (:callbacks xt:n-callback
				  (lambda (w)
				    (query "Quit OK?" (lambda (w) (quit))))))
	     )))

(define file-open-w
  (widget "file-open-w" xt:transient-shell
	   (:instantiation xt:create-popup-shell)
	   (:children
	     (lambda (parent)
	       (fsb parent
		    (lambda (s)
		      (xt:popdown (file-open-w #t))
		      (open-text-file s))
		    (lambda () (xt:popdown (file-open-w #t))))))
	   (:init (lambda (self child)
		    (center self)))))

;; TEXT-PANE
(define text-pane
  (widget "text-pane" xaw:ascii-text
	   (:resources xt:n-display-caret #f
		       xt:n-preferred-pane-size 200
		       xt:n-scroll-horizontal xaw:text-scroll-when-needed
		       xt:n-scroll-vertical xaw:text-scroll-always)))

(define (open-text-file filename)
  (let ((src ((widget "asciisrc" xaw:multi-src-object
		      (:resources xt:n-type xaw:ascii-file
				  xt:n-string filename))
	      toplevel-w)))
    (set! *editing-file* src)
    (xaw:text-set-source (text-pane #f) src 0)
    (xaw:text-display-caret (text-pane #f) #t)
    (xt:set-values (text-pane #f) xt:n-edit-type xaw:text-edit)))

;; FILE SELECTION BOX
;;
;; FSB PARENT APPLY-PROC CANCEL-PROC
;;   Create file selection box widget as PARENT's child.
;;   APPLY-PROC is a procedure which will be called with selected filename
;;   when 'apply' button is activated.  CANCEL-PROC is a procedure which
;;   will be called with no arguments.

(define (fsb parent apply-proc cancel-proc)
  ;; child widgets
  (define fsb-selection
    (widget "selection" xaw:ascii-text
	     (:resources xt:n-edit-type xaw:text-edit
			 xt:n-length 30
			 xt:n-resize xaw:text-resize-width
			 xt:n-skip-adjust #t)
	     (:init (lambda (self child)
		      (let ((source (xaw:text-get-source self)))
			(xt:add-callback source xt:n-callback cr-handler))))))
  (define fsb-list
    (widget "list" xaw:list
	     (:resources xt:n-list (fsb-get-directory "")
			 xt:n-default-columns 1
			 xt:n-force-columns #t)
	     (:init (lambda (self child)
		      (xaw:add-callback self xaw:list-callback
			   (lambda (w item . fo)
			     (xt:set-values (fsb-selection #t)
				 xt:n-string item)))))))
  (define fsb-viewport
    (widget "viewport" xaw:viewport
	     (:resources xt:n-allow-vert #t  xt:n-force-bars #t
			 xt:n-preferred-pane-size 150
			 xt:n-skip-adjust #f)
	     (:children fsb-list)
	     (:init (lambda (self child)
		      (xt:set-values self xt:n-height 120)))))

  ;;get directory contents specified by PREFIX
  (define (fsb-get-directory prefix)
    (let* ((p (open-pipe (string-append "ls -aF " prefix) "r"))
	   (l (let loop ((item (read-line p)))
		(if (eof-object? item) '()
		    (cons (string-append prefix item)
			  (loop (read-line p)))))))
      (close-port p)
      (list->vector l)))

  ;; deal with carriage return in the selection window
  (define (cr-handler w)
    (let* ((s (xt:get-value w xt:n-string xt:string))
	   (len (string-length s)))
      (if (and (> len 0)
	       (char=? (string-ref s (- len 1)) #\newline))
	  (begin
	    (xt:set-values w xt:n-string
		(substring s 0 (- len 1)))
	    (fsb-select)))))

  ;; deal with `apply'
  (define (fsb-select)
    (let ((selected (xt:get-value (fsb-selection #f)
			xt:n-string xt:string)))
      (if (= (string-length selected) 0)
	  '()
	  (case (string-ref selected (- (string-length selected) 1))
	    ((#\/)
	     (xaw:list-change (fsb-list #t) 
		  (fsb-get-directory selected) #f #t))
	    ((#\@ #\*)
	     (apply-proc (substring selected 0
				    (- (string-length selected) 1))))
	    (else
	     (apply-proc selected))))))

  ;; main body of fsb
  ((widget "fsb" xaw:paned
	    (:children
	      (widget "label" xaw:label
		       (:resources xt:n-label "Selection:"
				   xt:n-border-width 0
				   xt:n-skip-adjust #t))
	      fsb-selection
	      fsb-viewport
	      (widget "commands" xaw:box
		       (:resources xt:n-orientation "horizontal"
				   xt:n-skip-adjust #t)
		       (:children
			 (widget "apply" xaw:command
				  (:resources xt:n-label "Apply")
				  (:callbacks xt:n-callback
					      (lambda (w) (fsb-select))))
			 (widget "cancel" xaw:command
				  (:resources xt:n-label "Cancel")
				  (:callbacks xt:n-callback
					      (lambda (w) (cancel-proc)))))))
	    (:init (lambda (self children)
		     (map (lambda (w) (xt:set-values w xt:n-show-grip #f))
			  children))))
   parent))

;;
;; Some useful subroutines
;;

;; Place popup widget W at the center of toplevel widget.
(define (center w)
  (let ((top-x (xt:get-value toplevel-w xt:n-x xt:short))
	(top-y (xt:get-value toplevel-w xt:n-y xt:short))
	(top-w (xt:get-value toplevel-w xt:n-width xt:short))
	(top-h (xt:get-value toplevel-w xt:n-height xt:short))
	(pop-w (xt:get-value w xt:n-width xt:short))
	(pop-h (xt:get-value w xt:n-height xt:short)))
    (xt:set-values w
	xt:n-x (inexact->exact (+ top-x (/ (- top-w pop-w) 2)))
	xt:n-y (inexact->exact (+ top-y (/ (- top-h pop-h) 2))))))

;; Popup warning box with message MSG.

(define (warn msg)
  (letrec ((warn-dialog
	    (widget "warn-dialog" xaw:dialog
		     (:resources xt:n-label "")
		     (:children
		       (widget "ok" xaw:command
				(:resources xt:n-label "OK")
				(:callbacks xt:n-callback
					    (lambda (w)
					      (xt:popdown
						  (warn-shell toplevel-w))))
				))))
	   (warn-shell
	    (widget "warning" xt:transient-shell
		     (:instantiation xt:create-popup-shell)
		     (:children warn-dialog))))

    (xt:set-values (warn-dialog (warn-shell toplevel-w)) xt:n-label msg)
    (center (warn-shell toplevel-w))
    (xt:popup (warn-shell toplevel-w) 2)))

;; Popup dialog box with message MSG, with OK and CANCEL buttons.
;; If OK button is activated, callback procedure OK-PROC is invoked.
;; Popdown won't be done for OK button.

(define (query msg ok-proc)
  (letrec ((query-dialog
	    (widget "query-dialog" xaw:dialog
		     (:resources xt:n-label "")
		     (:children
		       (widget "ok" xaw:command
				(:resources xt:n-label "OK")
				(:callbacks xt:n-callback ok-proc))
		       (widget "cancel" xaw:command
				(:resources xt:n-label "Cancel")
				(:callbacks xt:n-callback
					    (lambda (w)
					      (xt:popdown
						  (query-shell toplevel-w))))
				))))
	   (query-shell
	    (widget "query" xt:transient-shell
		     (:instantiation xt:create-popup-shell)
		     (:children query-dialog))))

    (xt:set-values (query-dialog (query-shell toplevel-w))
	xt:n-label msg)
    (center (query-shell toplevel-w))
    (xt:popup (query-shell toplevel-w) 2)))


;; cliche
(define toplevel-w
  (if (defined? vs:top-level)
      (xt:app-create-shell "shell" "Shell"
			   xt:application-shell
			   (xt:display vs:top-level))
      (xt:initialize "xawsample" "XAwsample")))

(xt:set-values toplevel-w
	       xt:n-title "Scheme with Athena Widget")

(main-window toplevel-w)
(xt:realize-widget toplevel-w)

(xt:main-loop #t)
