;;;
;;;  sample for xawscm and widget macro package.
;;;
;;;  @(#)xawsample.scm 1.0.0 94/02/15 (Shiro Kawai, kawai@sail.t.u-tokyo.ac.jp)
;;;
;;(require 'macros-that-work)
(require 'repl)
(require (in-vicinity (library-vicinity) "x11.scm"))
(require (in-vicinity (library-vicinity) "xt.scm"))
(require (in-vicinity (library-vicinity) "xaw.scm"))
(require (in-vicinity (library-vicinity) "xevent.scm"))
(load  "widget")

;; MAIN-WINDOW
(define main-window
  (widget "main-window" xaw:paned
	  (:children command-box menu-pane)))

;; MENUS
;;  This is a sample of form and list widget.
(define menu-pane
  (widget "menu-pane" xaw:form
	  (:children
	   (widget "food-label" xaw:label
		   (:resources xt:n-label "Foods"
			       xt:n-border-width 0))
	   (widget "food-list" xaw:list
		   (:resources xt:n-list
			       #("Hamburgar" "Cheeseburgar"
						    "Meat pie" "Apple pie")
			       xt:n-default-columns 1
			       xt:n-vertical-list #t)
		   (:init (lambda (self child)
			    (xaw:add-callback self xaw:list-callback
					      list-select-callback)
			    (set! *food-list-w* self))))
	   (widget "drink-label" xaw:label
		   (:resources xt:n-label "Drinks"
			       xt:n-border-width 0))
	   (widget "drink-list" xaw:list
		   (:resources xt:n-list 
			       #("Coffee" "Tea" "Milk")
			       xt:n-default-columns 1
			       xt:n-vertical-list #t)
		   (:init (lambda (self child)
			    (xaw:add-callback self xaw:list-callback
					      list-select-callback)
			    (set! *drink-list-w* self)))))
	  (:init (lambda (self children)
		   (xt:set-values (cadr children)
				  xt:n-from-vert (car children))
		   (xt:set-values (caddr children)
				  xt:n-from-horiz (cadr children))
		   (xt:set-values (cadddr children)
				  xt:n-from-vert (caddr children)
				  xt:n-from-horiz (cadr children))))
	  ))

(define (list-select-callback w . args)
  (display "Item selected: ") (write args) (newline))

(define *food-list-w* '())
(define *drink-list-w* '())

;; COMMAND BOX
;;   This is a sample of box, command, repeater, and menu-button widget
(define command-box
  (widget "command-box" xaw:box
	  (:resources xt:n-border-width 0
		      xt:n-orientation "horizontal")
	  (:children
	   (widget "apply-button" xaw:repeater
		   (:resources xt:n-label "Apply"
			       xt:n-flash #t
			       xt:n-repeat-delay 500
			       xt:n-decay 30
			       xt:n-initial-delay 500)
		   (:callbacks xt:n-callback
			       (lambda (w)
				 (let ((food (xaw:list-show-current
					      *food-list-w*))
				       (drink (xaw:list-show-current
					       *drink-list-w*)))
				   (display "Order summary: ")
				   (display (if food (car food) "No food"))
				   (display ", ")
				   (display (if drink (car drink) "No drink"))
				   (newline)))))
	   (widget "cancel-button" xaw:command
		   (:resources xt:n-label "Cancel")
		   (:callbacks xt:n-callback
			       (lambda (w)
				 (xaw:list-unhighlight *food-list-w*)
				 (xaw:list-unhighlight *drink-list-w*))))
	   (widget "options-button" xaw:menu-button
		   (:resources xt:n-label "Options"
			       xt:n-menu-name "options-menu")
		   (:init (lambda (self child)
			    (options-menu self))))
	   (widget "quit-button" xaw:command
		   (:resources xt:n-label "Quit")
		   (:callbacks xt:n-callback
			       (lambda (w)
				 (xt:popup (quit-confirm toplevel-w) 0)
				 ))))
	  ))

;; OPTIONS MENU
;;  This is a sample of simple-menu widget
(define options-menu
  (widget "options-menu" xaw:simple-menu
	  (:instantiation xt:create-popup-shell)
	  (:resources xt:n-label "options")
	  (:children
	   (widget "" xaw:sme-line)
	   (widget "repl" xaw:sme-bsb
		   (:resources xt:n-label "interpreter")
		   (:callbacks xt:n-callback
			       (lambda (w)
				 (display "*** Internal interpreter invoked")
				 (newline)
				 (display "      use (repl:quit) to exit")
				 (newline)
				 (repl:top-level slib:eval)))))
	  ))

;; QUIT CONFIRMATION
;;   This is a sample of transient-shell and dialog widget.
(define quit-confirm
  (widget "Quit confirmation" xt:transient-shell
	  (:instantiation xt:create-popup-shell)
	  (:resources xt:n-x (+ 50
				(xt:get-value toplevel-w
					      xt:n-x xt:short))
		      xt:n-y (+ 50
				(xt:get-value toplevel-w
					      xt:n-y xt:short)))
	  (:children
	   (widget "dialog" xaw:dialog
		   (:resources xt:n-label "Really Quit?")
		   (:children
		    (widget "ok" xaw:command
			    (:resources xt:n-label "OK")
			    (:callbacks xt:n-callback
					(lambda (w) (quit))))
		    (widget "cancel" xaw:command
			    (:resources xt:n-label "Cancel")
			    (:callbacks xt:n-callback
					(lambda (w)
					  (xt:popdown (quit-confirm #f))))))
		   ))
	  ))

;; cliche
(define toplevel-w
  (if (defined? vs:top-level)
      (xt:app-create-shell "shell" "Shell"
			   xt:application-shell
			   (xt:display vs:top-level))
      (xt:initialize "xawsample" "XAwsample")))

(xt:set-values toplevel-w
	       xt:n-title "Scheme with Athena Widget")

(main-window toplevel-w)
(xt:realize-widget toplevel-w)

(xt:main-loop #t)
