/* $Id: strvec.c,v 1.5 1995/09/01 13:27:50 skimu Exp $
 */
#include "scm.h"
#include "strvec.h"

/* STRINGVECTOR, a C-array of strings terminated by NULL pointer.
   For List widget */

long tc16_stringvector;
static int print_stringvector P((SCM exp, SCM port, int writing));
static sizet free_stringvector P((CELLPTR ptr));
static smobfuns smobstringvector = {
  mark0, free_stringvector, print_stringvector, NULL
};

/* list->stringvector LIST-OF-STRINGS
   Return STRINGVECTOR object */

static char s_list2stringvector[] = "list->stringvector";
SCM list2stringvector(l)
     SCM l;
{
  int len = ilength(l);
  SCM new, s = l;
  char **table;
    
  ASSERT(len > 0, l, ARG1, s_list2stringvector);
  NEWCELL(new);
  DEFER_INTS;
  CAR(new) = tc16_stringvector;
/*
  table = (char **)must_malloc((len+1) * sizeof(char *),
			       s_list2stringvector);
  while (!NULLP(s))
    {
      SCM elt = CAR(s);
      ASSERT(NIMP(elt) && STRINGP(elt), l, "non-string element in ",
	     s_list2stringvector);
      *table = (char *)must_malloc( LENGTH(elt)+1, s_list2stringvector );
      strcpy(*table, CHARS(elt));
      table++;
      s = CDR(s);
    }
  *table = (char *)0;
*/
  table = makargvfrmstrs(l, s_list2stringvector);
  CDR(new) = (SCM)table;
  STRVECSETLENGTH(new, len);
  ALLOW_INTS;
  return new;
}

static char s_stringvector2list[] = "stringvector->list";
SCM stringvector2list(sv)
     SCM sv;
{
  ASSERT(NIMP(sv) && STRINGVECTORP(sv), sv, ARG1, s_stringvector2list);
  return makfromstrs(-1, (char **)CDR(sv));
}

static char s_stringvector_len[] = "stringvector-length";
SCM stringvector_len(sv)
     SCM sv;
{
  int i;
  char **p;
  ASSERT(NIMP(sv) && STRINGVECTORP(sv), sv, ARG1, s_stringvector_len);
/*
  p = CDR(sv);
  for (i=0; p[i]; i++);
  return MAKINUM(i);
*/
  return MAKINUM(STRVECLENGTH(sv));
}

static char s_stringvector_ref[] = "stringvector-ref";
SCM stringvector_ref(sv, k)
 SCM sv, k;
{
  char **p;
  int i;

  ASSERT(NIMP(sv) && STRINGVECTORP(sv), sv, ARG1, s_stringvector_ref);
  ASSERT(INUMP(k), k, ARG2, s_stringvector_ref);
  ASSERT(INUM(k) < STRVECLENGTH(sv) && (INUM(k) >= 0),
	 k, OUTOFRANGE, s_stringvector_ref);
  p = (char **)CDR(sv);
  i = INUM(k);
  return makfromstr(p[i], strlen(p[i]));
}

static char s_stringvector_set[] = "stringvector-set!";
SCM stringvector_set(sv, k, str)
     SCM sv, k, str;
{
  char **p;
  char *tmp;
  int i;
  ASSERT(NIMP(sv) && STRINGVECTORP(sv), sv, ARG1, s_stringvector_set);
  ASSERT(INUMP(k), k, ARG2, s_stringvector_set);
  ASSERT(NIMP(str) && STRINGP(str), str, ARG3, s_stringvector_set);
  ASSERT(INUM(k) < STRVECLENGTH(sv) && (INUM(k) >= 0),
	 k, OUTOFRANGE, s_stringvector_set);
  p = (char **)CDR(sv);
  i = INUM(k);
  free (p[i]);
  p[i] = must_malloc(LENGTH(str) + 1, s_stringvector_set);
  strcpy(p[i], CHARS(str));
  return UNSPECIFIED;
}

static char s_stringvector[] = "stringvector";
SCM stringvector(args)
     SCM args;
{
  return list2stringvector(args);
}

static char s_stringvectorp[] = "stringvector?";
SCM stringvectorp(s)
     SCM s;
{
  if (NIMP(s) && STRINGVECTORP(s))
    return BOOL_T;
  else
    return BOOL_F;
}

static sizet free_stringvector(ptr)
     CELLPTR ptr;
{
    int size;
    char **p = (char **)CDR(ptr);
    
    size = (STRVECLENGTH(ptr) + 1) * sizeof(char *);
    while (*p)
    {
	size += strlen(*p) + 1;
	p++;
    }
    must_free_argv((char **)CDR(ptr));
    return size;
}

static int print_stringvector(exp, f, writing)
     SCM exp;
     SCM f;
     int writing;
{
    char **p = (char **)CDR(exp);
    lputs("#<stringvector:", f);
    while (*p)
    {
	lputc(' ', f);
	lputc('"', f);
	lputs(*p, f);
	lputc('"', f);
	p++;
    }    lputc('>',f);
    return 1;
}

iproc strvec_subr1s[] = {
  {s_list2stringvector, list2stringvector},
  {s_stringvectorp,     stringvectorp},
  {s_stringvector2list, stringvector2list},
  {s_stringvector_len,  stringvector_len},
  {0,0}
};
iproc strvec_subr2s[] = {
  {s_stringvector_ref,  stringvector_ref},
  {0,0}
};
iproc strvec_subr3s[] = {
  {s_stringvector_set,  stringvector_set},
  {0,0}
};
iproc strvec_lsubrs[] = {
  {s_stringvector, stringvector},
  {0,0}
};
init_strvec()
{
  init_iprocs(strvec_subr1s, tc7_subr_1);
  init_iprocs(strvec_subr2s, tc7_subr_2);
  init_iprocs(strvec_subr3s, tc7_subr_3);
  init_iprocs(strvec_lsubrs, tc7_lsubr);
  tc16_stringvector = newsmob(&smobstringvector);
  add_feature("stringvector");
}
