/*
 * $Id: xaw.c,v 1.7 1995/09/01 12:46:20 skimu Exp $
 * Class and function definitions for scm interface to Athena Widgets
 */
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

#include <X11/Xaw/AsciiSink.h>
#include <X11/Xaw/AsciiSrc.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Box.h>
/* #include <X11/Xaw/Clock.h> */
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Dialog.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Grip.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/List.h>
/* #include <X11/Xaw/Logo.h> */
/* #include <X11/Xaw/Mailbox.h> */
#include <X11/Xaw/MultiSink.h>
#include <X11/Xaw/MultiSrc.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Panner.h>
#include <X11/Xaw/Porthole.h>
#include <X11/Xaw/Repeater.h>
#include <X11/Xaw/Scrollbar.h>
#include <X11/Xaw/Simple.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/SmeBSB.h>
#include <X11/Xaw/SmeLine.h>
#include <X11/Xaw/StripChart.h>
/* #include <X11/Xaw/Text.h> */
#include <X11/Xaw/Toggle.h>
#include <X11/Xaw/Tree.h>
#include <X11/Xaw/Viewport.h>
/* #include <X11/Xaw/XawInit.h> */

#include "scm.h"
#include "xt.h"
#include "strvec.h"

#ifndef PAIR_P
#define PAIR_P(x) (!(IMP(x)) && CONSP(x))       /* check this!! */
#endif
#ifndef BOOL
#define BOOL NFALSEP
#endif

xt_widget_class_t xaw_widget_classes[] = {
    "xaw:ascii-sink-object",	&asciiSinkObjectClass,
    "xaw:ascii-src-object",	&asciiSrcObjectClass,
    "xaw:ascii-text",		&asciiTextWidgetClass,       
    "xaw:box",			&boxWidgetClass,             
/*  "xaw:clock",		&clockWidgetClass,            */
    "xaw:command",		&commandWidgetClass,         
    "xaw:dialog",		&dialogWidgetClass,          
    "xaw:form",			&formWidgetClass,            
    "xaw:grip",			&gripWidgetClass,            
    "xaw:label",		&labelWidgetClass,           
    "xaw:list",			&listWidgetClass,            
/*  "xaw:logo",			&logoWidgetClass,            */
/*  "xaw:mailbox",		&mailboxWidgetClass,         */
    "xaw:menu-button",		&menuButtonWidgetClass,      
    "xaw:multi-sink-object",    &multiSinkObjectClass,
    "xaw:multi-src-object",     &multiSrcObjectClass,
    "xaw:menu-button",		&menuButtonWidgetClass,      
    "xaw:paned",		&panedWidgetClass,           
    "xaw:panner",		&pannerWidgetClass,          
    "xaw:porthole",		&portholeWidgetClass,        
    "xaw:repeater",		&repeaterWidgetClass,        
    "xaw:scrollbar",		&scrollbarWidgetClass,       
    "xaw:simple",		&simpleWidgetClass,          
    "xaw:simple-menu",		&simpleMenuWidgetClass,      
    "xaw:sme-object",		&smeObjectClass,
    "xaw:sme-bsb-object",	&smeBSBObjectClass,
    "xaw:sme-line-object",	&smeLineObjectClass,
    "xaw:strip-chart",		&stripChartWidgetClass,      
/*    "xaw:text",			&textWidgetClass,     */
    "xaw:toggle",		&toggleWidgetClass,          
    "xaw:tree",			&treeWidgetClass,            
    "xaw:viewport",		&viewportWidgetClass
};

/*
 * Public functions
 */

/*
 *   Special treatment for callback procedures
 */

/* It may be problem that callback functions writen in scheme cannot
   obtain call_data field, since some callbacks uses that field to
   pass important data.  For example, XtNcallback of List widget passes
   a selected item by call_data field.  To treat it, I add special
   callbacks which can receive data in call_data as additional argument.
   For List widget, use xaw:add-callback instead of xt:add-callback.
   Xaw:add-callback remembers the special callback procedure
   chaning it on global variable xaw:*callback-list*, and add
   `interpreter' callback (written in C) to specified List widget.
   When XtNcallback of the List widget is activated, `interpreter'
   callback interprets its call_data argument, then searches 
   xaw:*callback-list* for registered scheme callback procedures
   and invoke them.

   Following function is general subroutine to support such callbacks. */

static char s_xaw_callback_list[] = "xaw:*callback-list*";
static SCM xaw_callback_list = EOL;

static void xaw_list_callback_interpreter();
static void xaw_report_callback_interpreter();
static void xaw_scrollproc_callback_interpreter();
static void xaw_jumpproc_callback_interpreter();

struct interpreter_table {
    char *name;
    void (*proc)();
} xaw_interpreter_table[] = {
   { "xaw:list-callback",	xaw_list_callback_interpreter},
   { "xaw:report-callback",	xaw_report_callback_interpreter},
   { "xaw:scrollproc-callback",	xaw_scrollproc_callback_interpreter},
   { "xaw:jumpproc-callback",	xaw_jumpproc_callback_interpreter},
   { (char *)NULL, (void (*)())NULL }
}; 

/* This is from xt.c, xt_callback_handler */
static void xaw_callback_handler(w, proc, data)
     Widget w;
     XtPointer proc, data;
{
    SCM sw;
    SCM sproc = (SCM) proc;

    sw = make_widget(w);
    apply(sproc, sw, listofnull);
}

/* XAW:ADD-CALLBACK W INTERPRETER PROC
     Register special callback PROC to widget W.  INTERPRETER must be
     one of "xaw:list-callback", "xaw:report-callback",
     "xaw:scrollproc-callback" or "xaw:jumpproc-callback". */

static char s_xaw_add_callback[] = "xaw:add-callback";
SCM xaw_add_callback(sw, sinterpreter, rest)
     SCM sw, sinterpreter, rest;
{
    SCM entry, sproc;
    struct interpreter_table *interprt = xaw_interpreter_table;

    ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1, s_xaw_add_callback);
    ASSERT(NIMP(sinterpreter) && STRINGP(sinterpreter),
	   sinterpreter, ARG2, s_xaw_add_callback);
    ASSERT(PAIR_P(rest), rest, ARG3, s_xaw_add_callback);
    sproc = CAR(rest);
    ASSERT(NIMP(sproc) && CLOSUREP(sproc), sproc, ARG3,
	   s_xaw_add_callback);

    if (xaw_callback_list == EOL)
    {
	DEFER_INTS;
	xaw_callback_list =
	    sysintern(s_xaw_callback_list, EOL);
	ALLOW_INTS;
    }

    for (; interprt->name != NULL; interprt++)
    {
	if (!strcmp(interprt->name, CHARS(sinterpreter)))
	    break;
    }
    
    if (interprt->proc == NULL)
    {
	/* Ordinaly callback, use XtAddCallback */
	XtAddCallback(WIDGET(sw), CHARS(sinterpreter),
		      xaw_callback_handler, (XtPointer)sproc);
    }
    else
    {
	entry = assq(sw, CDR(xaw_callback_list));
	if (FALSEP(entry))
	{	
	    CDR(xaw_callback_list) =
		cons(cons(sw, cons(sproc, EOL)), CDR(xaw_callback_list));
	    XtAddCallback(WIDGET(sw), XtNcallback, interprt->proc, NULL);
	}
	else
	{
	    for (entry = CDR(entry); NIMP(CDR(entry)); entry = CDR(entry))
		;
	    CDR(entry) = cons(sproc, EOL);
	}
    }
    return UNSPECIFIED;
}

/* Apply callbacks registered by xaw:add-callback.
   Called from callback interpreter. */

static void xaw_apply_callbacks(w, aux_data)
     Widget w;		/* widget which invoked callback */
     SCM aux_data;	/* SCM interpreted call_data */
{
    SCM entry = BOOL_F, l;

    for (l = CDR(xaw_callback_list); NIMP(l); l = CDR(l))
	if (WIDGET(CAR(CAR(l))) == w)
	{
	    entry = CAR(l);
	    break;
	}
    
    if (NFALSEP(entry))
    {
	SCM sw = CAR(entry);
	for (entry = CDR(entry); NIMP(entry); entry = CDR(entry))
	    apply(CAR(entry), sw, aux_data);
    }
}

/*
 *   DIALOG widget support
 */

/* XAW:DIALOG-GET-VALUE-STRING WIDGET
     Returns string held in text widget associated to the
     dialog widget WIDGET. */
static char s_xaw_dialog_get_value_string[] = "xaw:dialog-get-value-string";
SCM xaw_dialog_get_value_string(sw)
     SCM sw;
{
    char *val;
    
    ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1, s_xaw_dialog_get_value_string);
    val = XawDialogGetValueString( WIDGET(sw) );
    return makfromstr(val, strlen(val));
}

/*
 *   FORM widget support
 */

/* XAW:FORM-DO-LAYOUT WIDGET DO-LAYOUT
     If DO-LAYOUT is #f, disables layout of the form widget WIDGET.
     Otherwise, enables it. */
static char s_xaw_form_do_layout[] = "xaw:form-do-layout";
SCM xaw_form_do_layout(sw, sdo_layout)
     SCM sw, sdo_layout;
{
    ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1, s_xaw_form_do_layout);
    XawFormDoLayout( WIDGET(sw), BOOL(sdo_layout) );
    return UNSPECIFIED;
}

/*
 *    LIST widget support
 */

/* XAW:LIST-CHANGE W LIST &optional LONGEST RESIZE
   Change the list of strings to be displayed.  W is list-widget,
   and LIST is either a list of strings or a stringarray object.
   Optional LONGEST is integer to specify the length of the longest item
   in the list in pixels.  Omitting it or Specifying non-integer value for
   it make the List widget calculate the length.
   Optional boolean value RESIZE specifies whether the List widget
   should try to resize itself or not. */

static char s_xaw_list_change[] = "xaw:list-change";
SCM xaw_list_change(sw, slist, rest)
     SCM sw, slist, rest;
{
    int restlen = ilength(rest);
    int longest, resize;
    String *list;
    
    ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1, s_xaw_list_change);
    ASSERT(NIMP(slist) && (PAIR_P(slist) || STRINGVECTORP(slist)),
	   slist, ARG2, s_xaw_list_change);
    if (PAIR_P(slist))
	slist = list2stringvector(slist);
    list = (String *)(CDR(slist));

    if (restlen > 1)
	longest = INUMP(CAR(rest))? 0 : INUM(CAR(rest));
    if (restlen > 2)
	resize = ( CAR(CDR(rest)) == BOOL_T );

    XawListChange( WIDGET(sw), list, 0, longest, resize);
    return UNSPECIFIED;
}

/* Special callback interpreter to make XawListReturnStruct available. */

static void xaw_list_callback_interpreter(w, client_data, call_data)
     Widget w;
     XtPointer client_data, call_data;
{
    XawListReturnStruct *selected = (XawListReturnStruct *)call_data;
    SCM string = makfromstr(selected->string,
			    (size_t)strlen(selected->string));
    SCM index = MAKINUM(selected->list_index);

    xaw_apply_callbacks(w, cons(string, cons(index, EOL)));
}

/* XAW:LIST-HIGHLIGHT W ITEM
   Highlight ITEM-th item in the List widget W.   Only one item can
   be highlightened at a time. */
static char s_xaw_list_highlight[] = "xaw:list-highlight";
SCM xaw_list_highlight(sw, sitem)
     SCM sw, sitem;
{
    ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1, s_xaw_list_highlight);
    ASSERT(INUMP(sitem), sw, ARG2, s_xaw_list_highlight);
    XawListHighlight(WIDGET(sw), INUM(sitem));
    return UNSPECIFIED;
}

/* XAW:LIST-UNHIGHLIGHT W
   Unhighlight the currently highlightened item in the List widget W. */
static char s_xaw_list_unhighlight[] = "xaw:list-unhighlight";
SCM xaw_list_unhighlight(sw)
     SCM sw;
{
    ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1, s_xaw_list_unhighlight);
    XawListUnhighlight(WIDGET(sw));
    return UNSPECIFIED;
}

/* XAW:LIST-SHOW-CURRENT W
   Return a pair of <label> and <index> of an item currently selected
   in the List widget W.  If no items is selected, #f is returned. */
static char s_xaw_list_show_current[] = "xaw:list-show-current";
SCM xaw_list_show_current(sw)
     SCM sw;
{
    XawListReturnStruct *selected;
    SCM r;
    ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1, s_xaw_list_show_current);
    selected = XawListShowCurrent(WIDGET(sw));
    if (selected->list_index == XAW_LIST_NONE)
	r = BOOL_F;
    else
    {
	SCM sstring = makfromstr(selected->string, strlen(selected->string));
	r = cons(sstring, MAKINUM(selected->list_index));
    }
    XtFree((char *)selected);
    return r;
}


/*
 *   PANED widget support
 */

/* XAW:PANED-ALLOW-RESIZE WIDGET ALLOW-RESIZE
     Disables (when ALLOW-RESIZE == #f) or enables (otherwise) a pane window
     WIDGET (a child of paned widget) for resizing requests. */
static char s_xaw_paned_allow_resize[] = "xaw:paned-allow-resize";
SCM xaw_paned_allow_resize(sw, sbool)
     SCM sw, sbool;
{
    ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1, s_xaw_paned_allow_resize);
    XawPanedAllowResize(WIDGET(sw), BOOL(sbool));
    return UNSPECIFIED;
}

/* XAW:PANED-SET-MIN-MAX WIDGET MIN MAX
     Set minimum and maximun height for a pane WIDGET (a child of paned
     widget). */
static char s_xaw_paned_set_min_max[] = "xaw:paned-set-min-max";
SCM xaw_paned_set_min_max(sw, smin, rest)
     SCM sw, smin, rest;
{
    ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1, s_xaw_paned_set_min_max);
    ASSERT(INUMP(smin), smin, ARG2, s_xaw_paned_set_min_max);
    ASSERT(PAIR_P(rest), rest, ARG3, s_xaw_paned_set_min_max);
    rest = CAR(rest);
    ASSERT(INUMP(rest), rest, ARG3, s_xaw_paned_set_min_max);
    XawPanedSetMinMax(WIDGET(sw), INUM(smin), INUM(rest));
    return UNSPECIFIED;
}

/* XAW:PANED-GET-MIN-MAX WIDGET
     Return pair of minimum and maximum height for a pane WIDGET (a child
     of paned widget). */
static char s_xaw_paned_get_min_max[] = "xaw:paned-get-min-max";
SCM xaw_paned_get_min_max(sw)
     SCM sw;
{
    int min, max;
    ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1, s_xaw_paned_get_min_max);
    XawPanedGetMinMax(WIDGET(sw), &min, &max);
    return cons(MAKINUM(min), MAKINUM(max));
}

/* XAW:PANED-SET-REFIGURE-MODE WIDGET MODE
     Enables (MODE != #f) or disables (otherwise) refiguration of
     a paned widget WIDGET */
static char s_xaw_paned_set_refigure_mode[] = "xaw:paned-set-refigure-mode";
SCM xaw_paned_set_refigure_mode(sw, smode)
     SCM sw, smode;
{
    ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1, s_xaw_paned_set_refigure_mode);
    XawPanedSetRefigureMode(WIDGET(sw), BOOL(smode));
    return UNSPECIFIED;
}

/* XAW:PANED-GET-NUM-SUB WIDGET
     Returns the number of panes in a paned widget WIDGET. */
static char s_xaw_paned_get_num_sub[] = "xaw:paned-get-num-sub";
SCM xaw_paned_get_num_sub(sw)
     SCM sw;
{
    int num;
    ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1, s_xaw_paned_get_num_sub);
    num = XawPanedGetNumSub(WIDGET(sw));
    return MAKINUM(num);
}

/*
 *   PANNER, PORTHOLE, VIEWPORT widget support
 */

/* Interpret XawPannerReport structure */

static void xaw_report_callback_interpreter(w, client_data, call_data)
     Widget w;
     XtPointer client_data, call_data;
{
    XawPannerReport *report = (XawPannerReport *)call_data;
    SCM entry = BOOL_F, l;
    SCM sreport = cons(MAKINUM(report->changed),
		       cons(cons(MAKINUM(report->slider_x),
				 MAKINUM(report->slider_y)),
			    cons(cons(MAKINUM(report->slider_width),
				      MAKINUM(report->slider_height)),
				 cons(cons(MAKINUM(report->canvas_width),
					   MAKINUM(report->canvas_height)),
				      EOL))));
    xaw_apply_callbacks(w, sreport);
}

/*
 *   SCROLLBAR widget support
 */

/* XAW:SCROLLBAR-SET-THUMB W TOP SHOWN
     Moves the visible thumb of scrollbar widget W to position
     TOP (0.0 -- 1.0) and length SHOWN (0.0 -- 1.0). */
static char s_xaw_scrollbar_set_thumb[] = "xaw:scrollbar-set-thumb";
SCM xaw_scrollbar_set_thumb(sw, top, rest)
     SCM sw, top, rest;
{
    float toppos, shown;
    
    ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1, s_xaw_scrollbar_set_thumb);
    ASSERT(INUMP(top) || (NIMP(top) && INEXP(top)), top, ARG2,
	   s_xaw_scrollbar_set_thumb);
    ASSERT(PAIR_P(rest), rest, ARG3, s_xaw_scrollbar_set_thumb);
    rest = CAR(rest);
    ASSERT(INUMP(rest) || (NIMP(rest) && INEXP(rest)), rest, ARG3,
	   s_xaw_scrollbar_set_thumb);
    toppos = INUMP(top)? (float)(INUM(top)) : (float)(REAL(top));
    shown = INUMP(rest)? (float)(INUM(rest)) : (float)(REAL(rest));
    XawScrollbarSetThumb(WIDGET(sw), toppos, shown);
    return UNSPECIFIED;
}

/* Interpret scrollProc and jumpProc callback */

static void xaw_scrollproc_callback_interpreter(w, client_data, position)
     Widget w;
     XtPointer client_data;
     int position;
{
    xaw_apply_callbacks(w, MAKINUM(position));
}

static void xaw_jumpproc_callback_interpreter(w, client_data, percent_ptr)
     Widget w;
     XtPointer client_data;
     XtPointer percent_ptr;
{
    double percent = (double)*(float *)percent_ptr;
    xaw_apply_callbacks(w, makdbl(percent, (double)0.0));
}

/*
 *   STRIPCHART widget support
 */

/* You cannot use XtNgetValue resource from scheme.  Use following instead:

   XAW:STRIP-CHART-SET-GETVALUE-PROC W PROC
   Set PROC as getValue procedure of StripChart widget W.
   PROC will be invoked on evely XtNupdate seconts, with W as the only
   argument.  PROC must return number. */

static void xaw_strip_chart_getvalue_handler(w, client_data, call_data)
     Widget w;
     XtPointer client_data;
     XtPointer call_data;
{
    SCM sproc = (SCM)client_data;
    double *resultptr = (double *)call_data;
    SCM value;
    
    value = apply(sproc, make_widget(w), EOL);
    
    if (INUMP(value))
	*resultptr = (double)INUM(value);
    else if (NIMP(value) && INEXP(value))
	*resultptr = (double)REALPART(value);
    else
    {
	*resultptr = 0.0;
	wta(value, "Illegal value is returned from getValue callback",
	    XtName(w));
    }
}

static char s_xaw_strip_chart_set_getvalue_proc[] =
    "xaw:strip-chart-set-getvalue-proc";
SCM xaw_strip_chart_set_getvalue_proc(sw, sproc)
     SCM sw, sproc;
{
    ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1,
	   s_xaw_strip_chart_set_getvalue_proc);
    ASSERT(NIMP(sproc) && CLOSUREP(sproc), sproc, ARG2,
	   s_xaw_strip_chart_set_getvalue_proc);
    
    /* First, remove previous XtNgetValue callback since
       it must contain just one callback procedure. */
    XtRemoveAllCallbacks(WIDGET(sw), XtNgetValue);
    
    XtAddCallback(WIDGET(sw), XtNgetValue,
		  xaw_strip_chart_getvalue_handler, (XtPointer)sproc);
    return UNSPECIFIED;
}

/*
 *   TEXT, ASCIITEXT widget support
 */

/* XAW:TEXT-SET-SELECTION W LEFT RIGHT
   Makes the text between LEFT and RIGHT primary selection. */
static char s_xaw_text_set_selection[] = "xaw:text-set-selection";
SCM xaw_text_set_selection(sw, sleft, rest)
     SCM sw, sleft, rest;
{
    SCM sright;
    
    ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1, s_xaw_text_set_selection);
    ASSERT(INUMP(sleft), sleft, ARG2, s_xaw_text_set_selection);
    ASSERT(PAIR_P(rest), rest, ARG3, s_xaw_text_set_selection);
    sright = CAR(rest);
    ASSERT(INUMP(sright), sright, ARG3, s_xaw_text_set_selection);
    
    XawTextSetSelection(WIDGET(sw), INUM(sleft), INUM(sright));
    return UNSPECIFIED;
}

/* XAW:TEXT-UNSET-SELECTION W
   Unhighlight previously highlighted text. */
static char s_xaw_text_unset_selection[] = "xaw:text-unset-selection";
SCM xaw_text_unset_selection(sw)
     SCM sw;
{
    ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1, s_xaw_text_unset_selection);
    XawTextUnsetSelection(WIDGET(sw));
    return UNSPECIFIED;
}

/* XAW:TEXT-GET-SELECTION-POS W
   Return pair of begin and end positions of selected text */
static char s_xaw_text_get_selection_pos[] = "xaw:text-get-selection-pos";
SCM xaw_text_get_selection_pos(sw)
     SCM sw;
{
    XawTextPosition begin, end;
    
    ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1, s_xaw_text_unset_selection);
    XawTextGetSelectionPos(WIDGET(sw), &begin, &end);
    return cons(MAKINUM((int)begin), MAKINUM((int)end));
}

/* XAW:TEXT-REPLACE W START END STRING
   Delete text between positions START and END, and insert STRING.
   Return value is integer, which may be compared with xaw:edit-done
   or xaw:edit-error (see xaw.scm) */
static char s_xaw_text_replace[] = "xaw:text-replace";
SCM xaw_text_replace(sw, sstart, rest)
     SCM sw, sstart, rest;
{
    XawTextPosition start, end;
    XawTextBlock block;
    int result;
    
    ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1, s_xaw_text_replace);
    ASSERT(INUMP(sstart), sstart, ARG2, s_xaw_text_replace);
    start = INUM(sstart);
    ASSERT(PAIR_P(rest), rest, ARG3, s_xaw_text_replace);
    end = CAR(rest);
    ASSERT(INUMP(end), end, ARG3, s_xaw_text_replace);
    end = INUM(end);
    rest = CDR(rest);
    ASSERT(PAIR_P(rest), rest, ARG4, s_xaw_text_replace);
    rest = CAR(rest);
    ASSERT(NIMP(rest) && STRINGP(rest), rest, ARG4, s_xaw_text_replace);
    
    block.ptr = CHARS(rest);
    block.firstPos = 0;
    block.length = strlen(block.ptr);
    block.format = FMT8BIT;
    
    result = XawTextReplace(WIDGET(sw), start, end, &block);
    return MAKINUM(result);
}

/* XAW:TEXT-SEARCH W DIRECTION STRING
   Search for a string STRING in th Text widget W.  Specify either
   xaw:sd-left or xaw:sd-right for DIRECTION.
   If the STRING is found, the location of the first character is returned.
   If search is failed, xaw:text-search-error is returned. */
static char s_xaw_text_search[] = "xaw:text-search";
SCM xaw_text_search(sw, sdir, rest)
     SCM sw, sdir, rest;
{
    XawTextPosition result;
    XawTextBlock block;
    
    ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1, s_xaw_text_search);
    ASSERT(INUMP(sdir), sdir, ARG2, s_xaw_text_search);
    ASSERT(PAIR_P(rest), rest, ARG3, s_xaw_text_search);
    rest = CAR(rest);
    ASSERT(NIMP(rest) && STRINGP(rest), rest, ARG3, s_xaw_text_search);

    block.firstPos = 0;
    block.length = LENGTH(rest);
    block.ptr = CHARS(rest);
    block.format = FMT8BIT;
    
    result = XawTextSearch(WIDGET(sw), INUM(sdir), &block);
    return MAKINUM(result);
}

/* XAW:TEXT-INVALIDATE W FROM TO
   Redisplay characters in the range from FROM to TO. */
static char s_xaw_text_invalidate[] = "xaw:text-invalidate";
SCM xaw_text_invalidate(sw, sfrom, rest)
     SCM sw, sfrom, rest;
{
    ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1, s_xaw_text_search);
    ASSERT(INUMP(sfrom), sfrom, ARG2, s_xaw_text_search);
    ASSERT(PAIR_P(rest), rest, ARG3, s_xaw_text_search);
    rest = CAR(rest);
    ASSERT(INUMP(rest), rest, ARG3, s_xaw_text_search);
    
    XawTextInvalidate(WIDGET(sw), INUM(sfrom), INUM(rest));
    return UNSPECIFIED;
}

/* XAW:TEXT-ENABLE-REDISPLAY W
   Flushes any changed due to batched updates when xaw:text-disable-redisplay.
   */
static char s_xaw_text_enable_redisplay[] = "xaw:text-enable-redisplay";
SCM xaw_text_enable_redisplay(sw)
     SCM sw;
{
    ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1,
	   s_xaw_text_enable_redisplay);
    XawTextEnableRedisplay( WIDGET(sw) );
    return UNSPECIFIED;
}

/* XAW:TEXT-DISABLE-REDISPLAY W
   Disable redisplay. */
static char s_xaw_text_disable_redisplay[] = "xaw:text-disable-redisplay";
SCM xaw_text_disable_redisplay(sw)
     SCM sw;
{
    ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1,
	   s_xaw_text_disable_redisplay);
    XawTextDisableRedisplay( WIDGET(sw) );
    return UNSPECIFIED;
}

/* XAW:TEXT-DISPLAY W
   Display any batched updates. */
static char s_xaw_text_display[] = "xaw:text-display";
SCM xaw_text_display(sw)
     SCM sw;
{
    ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1,
	   s_xaw_text_display);
    XawTextDisplay( WIDGET(sw) );
    return UNSPECIFIED;
}

/* XAW:TEXT-TOP-POSITION W
   Return the character position of the leftmost character on the first
   line displayed in the widget W. */
static char s_xaw_text_top_position[] = "xaw:text-top-position";
SCM xaw_text_top_position(sw)
     SCM sw;
{
    XawTextPosition result;
    
    ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1,
	   s_xaw_text_top_position);
    result = XawTextTopPosition( WIDGET(sw) );
    return MAKINUM((int)result);
}

/* XAW:TEXT-SET-SELECTION-ARRAY W . SARRAY
   New selection array SARRAY to a text widget W.
   SARRAY is a list of selection types, one of following values,
   defined in xaw.scm:
     xaw:select-null, xaw:select-position, xaw:select-char,
     xaw:select-word, xaw:select-line, xaw:select-paragraph,
     xaw:select-all
   Note that list may or may not terminated by xaw:select-null, though
   Xlib function of C requires the array is terminated by it.
   This stub function adds XawselectNull if it is absent. */
#define TextSelectArraySize	16

static char s_xaw_text_set_selection_array[] = "xaw:text-set-selection-array";
SCM xaw_text_set_selection_array(args)
     SCM args;
{
    int alen;
    XawTextSelectType sarray[TextSelectArraySize], *parray, *p;
    SCM sw, rest;
    
    ASSERT(PAIR_P(args), args, ARG1, s_xaw_text_set_selection_array);
    sw = CAR(args);
    rest = CDR(args);
    ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1,
	   s_xaw_text_set_selection_array);
/*    alen = ilength(sarray); */
    alen = ilength(rest);
    if (alen > 0)
    {
	if (alen < TextSelectArraySize)
	    p = parray = sarray;
	else
	    p = parray =
		(XawTextSelectType *)
		    must_malloc(sizeof(XawTextSelectType) * (alen+1),
				s_xaw_text_set_selection_array);
	
	for (; PAIR_P(rest); rest = CDR(rest))
	{
	    ASSERT(INUMP(CAR(rest)), rest, "Rest arguments",
		   s_xaw_text_set_selection_array);
	    *p++ = (XawTextSelectType)INUM(CAR(rest));
	}
	*p = XawselectNull;
	
	XawTextSetSelectionArray(WIDGET(sw), parray);
	if (parray != sarray)
	    free(parray);
    }
    return UNSPECIFIED;
}

/* XAW:TEXT-SET-INSERTION-POINT W POSITION
   Move the insertion caret to the source position POSITION. */
static char s_xaw_text_set_insertion_point[] =
    "xaw:text-set-insertion-point";
SCM xaw_text_set_insertion_point(sw, spos)
     SCM sw, spos;
{
    ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1,
	   s_xaw_text_set_insertion_point);
    ASSERT(INUMP(spos), spos, ARG2, s_xaw_text_set_insertion_point);

    XawTextSetInsertionPoint(WIDGET(sw), (XawTextPosition)INUM(spos));
    return UNSPECIFIED;
}

/* XAW:TEXT-GET-INSERTION-POINT W
   Get the current position of the insertion caret. */
static char s_xaw_text_get_insertion_point[] =
    "xaw:text-get-insertion-point";
SCM xaw_text_get_insertion_point(sw)
     SCM sw;
{
    XawTextPosition pos;
    
    ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1,
	   s_xaw_text_get_insertion_point);

    pos = XawTextGetInsertionPoint( WIDGET(sw) );
    return MAKINUM((int)pos);
}

/* XAW:TEXT-SET-SOURCE W SOURCE POSITION
   Replace the text source in the specified widget.
   A display update will be performed unless redisplay is disabled. */
static char s_xaw_text_set_source[] = "xaw:text-set-source";
SCM xaw_text_set_source(sw, src, rest)
     SCM sw, src, rest;
{
    ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1, s_xaw_text_set_source);
    ASSERT(NIMP(src) && WIDGETP(src), src, ARG2, s_xaw_text_set_source);
    ASSERT(PAIR_P(rest), rest, ARG3, s_xaw_text_set_source);
    rest = CAR(rest);
    ASSERT(INUMP(rest), rest, ARG3, s_xaw_text_set_source);
    
    XawTextSetSource( WIDGET(sw), WIDGET(src), (XawTextPosition)INUM(rest) );
    return UNSPECIFIED;
}

/* XAW:TEXT-GET-SOURCE W
   Get current text source widget. */
static char s_xaw_text_get_source[] = "xaw:text-get-source";
SCM xaw_text_get_source(sw)
     SCM sw;
{
    Widget src;
    
    ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1,  s_xaw_text_get_source);

    src = XawTextGetSource( WIDGET(sw) );
    return make_widget(src);
}

/* XAW:TEXT-DISPLAY-CARET W VISIBLE
   Pass #f/other object to VISIBLE to disable/enable displaying caret. */
static char s_xaw_text_display_caret[] = "xaw:text-display-caret";
SCM xaw_text_display_caret(sw, svisible)
     SCM sw, svisible;
{
    ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1, s_xaw_text_display_caret);
    XawTextDisplayCaret( WIDGET(sw), NFALSEP(svisible) );
    return UNSPECIFIED;
}

/* Following procedures do not appear in O'Reilly's Reference (for R4/R5),
   but described in X11/Xaw/AsciiSrc.h */

static char s_xaw_ascii_source_free_string[] = "xaw:ascii-source-free-string";
SCM xaw_ascii_source_free_string(sw)
     SCM sw;
{
    ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1,
	   s_xaw_ascii_source_free_string);
    XawAsciiSourceFreeString( WIDGET(sw) );
    return UNSPECIFIED;
}

static char s_xaw_ascii_save[] = "xaw:ascii-save";
SCM xaw_ascii_save(sw)
     SCM sw;
{
    ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1, s_xaw_ascii_save);
    return XawAsciiSave( WIDGET(sw) ) == True? BOOL_T : BOOL_F;
}

static char s_xaw_ascii_save_as_file[] = "xaw:ascii-save-as-file";
SCM xaw_ascii_save_as_file(sw, name)
     SCM sw, name;
{
    ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1, s_xaw_ascii_save_as_file);
    ASSERT(NIMP(name) && STRINGP(name), name, ARG2, s_xaw_ascii_save_as_file);
    return XawAsciiSaveAsFile( WIDGET(sw), CHARS(name) ) == True?
	BOOL_T : BOOL_F;
}

static char s_xaw_ascii_source_changed[] = "xaw:ascii-source-changed";
SCM xaw_ascii_source_changed(sw)
     SCM sw;
{
    ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1, s_xaw_ascii_source_changed);
    return XawAsciiSourceChanged( WIDGET(sw) ) == True? BOOL_T : BOOL_F;
}

/*
 *   TOGGLE widget support
 */
/* Note for radioData resource:
   In C, you can set any type of object for radioData resource, which
   will be used to identify a currently set toggle button in a radio
   group.  In xscm, I restrict the type of radioData resource to
   string object. */

/* XAW:TOGGLE-CHANGE-RADIO-GROUP W RADIO-GROUP
   Change radio group of toggle widget W.  If RADIO-GROUP is an another
   toggle widget, W is added to the radio group of it.  If RADIO-GROUP is
   #f, W is removed from any radio group. */
static char s_xaw_toggle_change_radio_group[] =
    "xaw:toggle-change-radio-group";
SCM xaw_toggle_change_radio_group(sw, rg)
     SCM sw, rg;
{
    ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1,
	   s_xaw_toggle_change_radio_group);
    ASSERT(NIMP(rg) && WIDGETP(rg), rg, ARG2,
	   s_xaw_toggle_change_radio_group);
    XawToggleChangeRadioGroup( WIDGET(sw), WIDGET(rg) );
    return UNSPECIFIED;
}

/* XAW:TOGGLE-GET-CURRENT RADIO-GROUP
   Return radioData of the Toggle in this radio group that is currently set.
   RadioData is a string object (see above note), whose default value is
   the name of that Toggle widget.
   If no Toggle is set in RADIO-GROUP, #f is returned. */
static char s_xaw_toggle_get_current[] = "xaw:toggle-get-current";
SCM xaw_toggle_get_current(rg)
     SCM rg;
{
    XtPointer current;
    
    ASSERT(NIMP(rg) && WIDGETP(rg), rg, ARG1, s_xaw_toggle_get_current);
    current = XawToggleGetCurrent( WIDGET(rg) );
    if (current != NULL)
	return makfromstr((char *)current, strlen((char *)current));
    else
	return BOOL_F;
}

/* XAW:TOGGLE-SET-CURRENT RADIO-GROUP RADIO-DATA
   Change the currently set Toggle in the RADIO-GROUP.
   RADIO-DATA is a string, and the Toggle which has the same string
   (in eq? sense) in radioData resource is selected. */
static char s_xaw_toggle_set_current[] = "xaw:toggle-set-current";
SCM xaw_toggle_set_current(rg, rd)
     SCM rg, rd;
{
    ASSERT(NIMP(rg) && WIDGETP(rg), rg, ARG1, s_xaw_toggle_set_current);
    ASSERT(NIMP(rd) && STRINGP(rd), rd, ARG2, s_xaw_toggle_set_current);
    XawToggleSetCurrent( WIDGET(rg), CHARS(rd) );
    return UNSPECIFIED;
}

/* XAW:TOGGLE-UNSET-CURRENT RADIO-GROUP
   Unset all Toggle widgets in RADIO-GROUP. */
static char s_xaw_toggle_unset_current[] = "xaw:toggle-unset-current";
SCM xaw_toggle_unset_current(rg)
     SCM rg;
{
    ASSERT(NIMP(rg) && WIDGETP(rg), rg, ARG1, s_xaw_toggle_unset_current);
    XawToggleUnsetCurrent( WIDGET(rg) );
    return UNSPECIFIED;
}

/*
 *   TREE widget support
 */

/* XAW:TREE-FORCE-LAYOUT W
   Force a Tree widget to re-layout its children. */
static char s_xaw_tree_force_layout[] = "xaw:tree-force-layout";
SCM xaw_tree_force_layout(sw)
     SCM sw;
{
    ASSERT(NIMP(sw) && WIDGETP(sw), sw, ARG1, s_xaw_tree_force_layout);
    XawTreeForceLayout( WIDGET(sw) );
    return UNSPECIFIED;
}

/*
 *   Initialization
 */

iproc xaw_subr1s[] = {
    { s_xaw_dialog_get_value_string,	xaw_dialog_get_value_string },
    { s_xaw_list_unhighlight,		xaw_list_unhighlight },
    { s_xaw_list_show_current,		xaw_list_show_current },
    { s_xaw_paned_get_min_max,		xaw_paned_get_min_max }, 
    { s_xaw_paned_get_num_sub,		xaw_paned_get_num_sub },
    { s_xaw_text_unset_selection,	xaw_text_unset_selection },
    { s_xaw_text_get_selection_pos,	xaw_text_get_selection_pos }, 
    { s_xaw_text_enable_redisplay,	xaw_text_enable_redisplay}, 
    { s_xaw_text_disable_redisplay,	xaw_text_disable_redisplay},
    { s_xaw_text_display,		xaw_text_display},
    { s_xaw_text_top_position,		xaw_text_top_position},  
    { s_xaw_text_get_insertion_point,   xaw_text_get_insertion_point}, 
    { s_xaw_text_get_source,		xaw_text_get_source}, 
    { s_xaw_ascii_source_free_string,	xaw_ascii_source_free_string},
    { s_xaw_ascii_save,			xaw_ascii_save},
    { s_xaw_ascii_source_changed,	xaw_ascii_source_changed},
    { s_xaw_toggle_get_current,		xaw_toggle_get_current},
    { s_xaw_toggle_unset_current,	xaw_toggle_unset_current},
    { s_xaw_tree_force_layout,		xaw_tree_force_layout},
    { 0, 0 }
}; 

iproc xaw_subr2s[] = {
    { s_xaw_form_do_layout,		xaw_form_do_layout },
    { s_xaw_list_highlight,		xaw_list_highlight },
    { s_xaw_paned_allow_resize,		xaw_paned_allow_resize }, 
    { s_xaw_paned_set_refigure_mode,	xaw_paned_set_refigure_mode },
    { s_xaw_strip_chart_set_getvalue_proc,xaw_strip_chart_set_getvalue_proc },
    { s_xaw_text_set_insertion_point,	xaw_text_set_insertion_point},
    { s_xaw_text_display_caret,		xaw_text_display_caret}, 
    { s_xaw_ascii_save_as_file,		xaw_ascii_save_as_file},  
    { s_xaw_toggle_change_radio_group,	xaw_toggle_change_radio_group}, 
    { s_xaw_toggle_set_current,		xaw_toggle_set_current}, 
    { 0, 0 }
}; 

iproc xaw_lsubrs[] = {
    { s_xaw_text_set_selection_array,	xaw_text_set_selection_array},
    { 0, 0 }
};
 
iproc xaw_lsubr2s[] = {
/*    { s_xaw_list_change,		xaw_list_change }, */
    { s_xaw_paned_set_min_max,		xaw_paned_set_min_max }, 
    { s_xaw_scrollbar_set_thumb,	xaw_scrollbar_set_thumb },
    { s_xaw_add_callback,		xaw_add_callback },  
    { s_xaw_text_set_selection,		xaw_text_set_selection },
    { s_xaw_text_replace,		xaw_text_replace },  
    { s_xaw_text_search,		xaw_text_search},
    { s_xaw_text_invalidate,		xaw_text_invalidate}, 
    { s_xaw_text_set_source,		xaw_text_set_source},  
    { 0, 0 }
}; 
    
void init_xaw()
{
    init_iprocs(xaw_subr1s, tc7_subr_1);
    init_iprocs(xaw_subr2s, tc7_subr_2);
    init_iprocs(xaw_lsubrs, tc7_lsubr);
    init_iprocs(xaw_lsubr2s, tc7_lsubr_2);
    xt_init_widget_classes(xaw_widget_classes,
			   XtNumber(xaw_widget_classes),
			   "*athena-widget-classes*");	
}
